#
# Util for WebDriver
#

from string import Formatter
from functools import wraps

from .locator import Locator
from .keys import Keys


class MemorizeFormatter(Formatter):
    """Customize the Formatter to record used and unused kwargs."""

    def __init__(self):
        """Initialize the MemorizeFormatter."""
        Formatter.__init__(self)
        self._used_kwargs = {}
        self._unused_kwargs = {}

    def check_unused_args(self, used_args, args, kwargs):
        """Implement the check_unused_args in superclass."""
        for k, v in kwargs.items():
            if k in used_args:
                self._used_kwargs.update({k: v})
            else:
                self._unused_kwargs.update({k: v})

    def vformat(self, format_string, args, kwargs):
        """Clear used and unused dicts before each formatting."""
        self._used_kwargs = {}
        self._unused_kwargs = {}
        return super().vformat(format_string, args, kwargs)

    def format_map(self, format_string, mapping):
        """format a string by a map

        Args:
            format_string(str): A format string
            mapping(dict): A map to format the string

        Returns:
            A formatted string.

        Raises:
            KeyError: if key is not provided by the given map.
        """
        return self.vformat(format_string, args=None, kwargs=mapping)

    def get_used_kwargs(self):
        """Get used kwargs after formatting."""
        return self._used_kwargs

    def get_unused_kwargs(self):
        """Get unused kwargs after formatting."""
        return self._unused_kwargs


def add_element_extension_method(Klass):
    """Add element_by alias and extension' methods(if_exists/or_none)."""
    def add_element_method(Klass, using):
        locator = using.name.lower()
        find_element_name = "element_by_" + locator
        find_element_if_exists_name = "element_by_" + locator + "_if_exists"
        find_element_or_none_name = "element_by_" + locator + "_or_none"
        wait_for_element_name = "wait_for_element_by_" + locator

        find_elements_name = "elements_by_" + locator
        wait_for_elements_name = "wait_for_elements_by_" + locator

        def find_element(self, value):
            return self.element(using.value, value)

        find_element.__name__ = find_element_name
        find_element.__doc__ = (
            "Set parameter 'using' to '{0}'.\n".format(using.value) +
            str(Klass.__dict__['element'].__doc__)
        )

        def find_element_if_exists(self, value):
            return self.element_if_exists(using.value, value)

        find_element_if_exists.__name__ = find_element_if_exists_name
        find_element_if_exists.__doc__ = (
            "Set parameter 'using' to '{0}'.\n".format(using.value) +
            str(Klass.__dict__['element_if_exists'].__doc__)
        )

        def find_element_or_none(self, value):
            return self.element_or_none(using.value, value)

        find_element_or_none.__name__ = find_element_or_none_name
        find_element_or_none.__doc__ = (
            "Set parameter 'using' to '{0}'.\n".format(using.value) +
            str(Klass.__dict__['element_or_none'].__doc__)
        )

        def wait_for_element_by(self, *args, **kwargs):
            return self.wait_for_element(using.value, *args, **kwargs)

        wait_for_element_by.__name__ = wait_for_element_name
        wait_for_element_by.__doc__ = (
            "Set parameter 'using' to '{0}'.\n".format(using.value) +
            str(Klass.__dict__['wait_for_element'].__doc__)
        )

        def find_elements(self, value):
            return self.elements(using.value, value)

        find_elements.__name__ = find_elements_name
        find_elements.__doc__ = (
            "Set parameter 'using' to '{0}'.\n".format(using.value) +
            str(Klass.__dict__['elements'].__doc__)
        )

        def wait_for_elements_available(self, *args, **kwargs):
            return self.wait_for_elements(using.value, *args, **kwargs)

        wait_for_elements_available.__name__ = wait_for_elements_name
        wait_for_elements_available.__doc__ = (
            "Set parameter 'using' to '{0}'.\n".format(using.value) +
            str(Klass.__dict__['wait_for_elements'].__doc__)
        )

        setattr(Klass, find_element_name, find_element)
        setattr(Klass, find_element_if_exists_name, find_element_if_exists)
        setattr(Klass, find_element_or_none_name, find_element_or_none)
        setattr(Klass, wait_for_element_name, wait_for_element_by)
        setattr(Klass, find_elements_name, find_elements)
        setattr(Klass, wait_for_elements_name, wait_for_elements_available)

    for locator in iter(Locator):
        add_element_method(Klass, locator)


def fluent(func):
    """Fluent interface decorator to return self if method return None."""
    @wraps(func)
    def fluent_interface(instance, *args, **kwargs):
        ret = func(instance, *args, **kwargs)
        if ret is not None:
            return ret
        return instance
    return fluent_interface


def value_to_key_strokes(value):
    """Convert value to a list of key strokes
    >>> value_to_key_strokes(123)
    ['1', '2', '3']
    >>> value_to_key_strokes('123')
    ['1', '2', '3']
    >>> value_to_key_strokes([1, 2, 3])
    ['1', '2', '3']
    >>> value_to_key_strokes(['1', '2', '3'])
    ['1', '2', '3']

    Args:
        value(int|str|list)

    Returns:
        A list of string.
    """
    result = []
    if isinstance(value, int):
        value = str(value)

    for v in value:
        if isinstance(v, Keys):
            result.append(v.value)
        elif isinstance(v, int):
            result.append(str(v))
        else:
            result.append(v)
    return result



#
# https = //w3c.github.io/webdriver/webdriver-spec.html
# WebDriver Result
#


from .webdriverexception import WebDriverException, find_exception_by_code


class WebDriverResult:
    """WebDriver result object.

    Atrributes:
        session_id(str): A UDID used to uniquely identify each session.
        status(int): A status code summarizing the result of the command.
                     A non-zero value indicates that the command failed.
        value(str|list|dict): The response JSON value.
    """

    def __init__(self, session_id, status, value):
        self.session_id = session_id
        self.status = status
        self.value = value

    @classmethod
    def from_object(cls, obj):
        """The factory method to create WebDriverResult from JSON Object.

        Args:
            obj(dict): The JSON Object returned by server.
        """
        return cls(
            obj.get('sessionId', None),
            obj.get('status', 0),
            obj.get('value', None)
        )

    def raise_for_status(self):
        """Raise WebDriverException if returned status is not zero."""
        if not self.status:
            return

        error = find_exception_by_code(self.status)
        message = None
        screen = None
        stacktrace = None

        if isinstance(self.value, str):
            message = self.value
        elif isinstance(self.value, dict):
            message = self.value.get('message', None)
            screen = self.value.get('screen', None)
            stacktrace = self.value.get('stacktrace', None)

        raise WebDriverException(error, message, screen, stacktrace)



#
# Default Asserts for element
#


from .webdriverexception import WebDriverException


def is_displayed(target):
    """Assert whether the target is displayed

    Args:
        target(WebElement): WebElement Object.

    Returns:
        Return True if the element is displayed or return False otherwise.
    """
    is_displayed = getattr(target, 'is_displayed', None)
    if not is_displayed or not hasattr(is_displayed, '__call__'):
        raise TypeError('Target has no attribute \'is_displayed\' or not callable')
    if not is_displayed():
        raise WebDriverException('element not visible')


def is_not_displayed(target):
    """Assert whether the target is not displayed

    Args:
        target(WebElement): WebElement Object.

    Returns:
        Return True if the element is not displayed or return False otherwise.
    """
    is_displayed = getattr(target, 'is_displayed', None)
    if not is_displayed or not hasattr(is_displayed, '__call__'):
        raise TypeError('Target has no attribute \'is_displayed\' or not callable')
    if is_displayed():
        raise WebDriverException('element is visible')



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='wd',

    version='0.1.0',

    description='Macaca Python Client',
    long_description=long_description,

    url='https://github.com/macacajs/wd.py',

    author='Zichen Zhu',
    author_email='zic.zhu@gmail.com',

    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    keywords=[
        'macaca',
        'webdriver',
        'uitest',
        'mobile automation',
        'e2e'
    ],

    packages=find_packages(exclude=['tests*']),

    install_requires=[
        'requests',
        'retrying'
    ],

    extras_require={
        'test': ['pytest', 'tox', 'pytest-xdist', 'pytest-cov', 'coverage', 'responses']
    }
)
